#! /bin/sh

test -f summary.dat && cat summary.dat >> summary.bak
test -f control.dat && cat control.dat >> control.bak
test -f detail.dat  && mv -f detail.dat   detail.bak
test -f snp.err     && mv -f snp.err      snp.err.bak
test -f snp.out     && mv -f snp.out      snp.out.bak

cp null.dat summary.dat
cp null.dat detail.dat

rm -f core core.*

make >snp.out 2>&1  &&  ${PWD}/snp >>snp.out 2>snp.err

grep -i snp.err detail.dat > tmp
grep -v -i "standard error" tmp >> error
rm -f tmp

case $? in
  0) exit 0 ;;
  esac
exit 1;
