#! /bin/sh

test -f emm.err     && mv -f emm.err      emm.err.bak
test -f emm.out     && mv -f emm.out      emm.out.bak

rm -f core core.*

make >emm.out 2>&1  &&  ${PWD}/emm $1 >>emm.out 2>emm.err

case $? in
  0) exit 0 ;;
  esac
exit 1;
